/* 

  ****************   NO WARRANTY  *****************

Since the Aspirin/MIGRAINES system is licensed free of charge,
Russell Leighton and the MITRE Corporation provide absolutley 
no warranty. Should the Aspirin/MIGRAINES system prove defective, 
you must assume the cost of all necessary servicing, repair or correction.
In no way will Russell Leighton or the MITRE Corporation be liable to you for
damages, including any lost profits, lost monies, or other
special, incidental or consequential damages arising out of
the use or inability to use the Aspirin/MIGRAINES system.

  *****************   COPYRIGHT  *******************

This software is the copyright of Russell Leighton and the MITRE Corporation. 
It may be freely used and modified for research and development
purposes. We require a brief acknowledgement in any research
paper or other publication where this software has made a significant
contribution. If you wish to use it for commercial gain you must contact 
The MITRE Corporation for conditions of use. Russell Leighton and 
the MITRE Corporation provide absolutely NO WARRANTY for this software.

   August, 1992 
   Russell Leighton
   The MITRE Corporation
   7525 Colshire Dr.
   McLean, Va. 22102-3481

*/
#include "spiral.h"

float Sigmoid(float x) /* [-0.5,0.5] */
{
  return ( (1.0 / (1.0 + AM_EXP((double)-x))) - 0.5);
}

float SigmoidPrime(float x)
{
  x = Sigmoid(x);
  return ( 0.25 - (x * x));
}


float L2Error (float *target, float *output, float *credit, int n)
{
  float mag_squared = 0.0;
  int i;

  for(i=0;i<n;i++) {
    credit[i] = target[i] - output[i];
    mag_squared += credit[i] * credit[i];
  }/* end for */

  return( mag_squared / 2.0 );
}


float L1Error (float *target, float *output, float *credit, int n)
{
  float mag = 0.0;
  int i;

  for(i=0;i<n;i++) {
    float val, diff;

    diff = target[i] - output[i] ;

    if (diff == 0.0)
      val = 0.0 ;
    else
      val = ( diff < 0.0 ) ? -1.0 : 1.0 ;

    credit[i] = val;
    mag += AM_FABS(diff);
  }/* end for */

  return( mag );
}



/*
**----------------------------------------------------------------------
**
**  Asymmetric thresholded squared error
**
**----------------------------------------------------------------------
*/

#define DELTA 0.150

#define GI 1.384   /* Credit multiplier for an Inhibited node */
#define GA 1.384   /* Credit multiplier for an Active node */

#define GIdb2 0.692
#define GAdb2 0.692

float asymmError(float *target, float *output, float *credit, int nx)
{
  float e;
  double t, total_error = 0.0;
  unsigned int k;
  
  for (k=0; k<nx; ++k) {
    
    e = output[k] - target[k];    
    
    if ( e < -DELTA ) {
      t = e + DELTA;
      credit[k] = -GA * t;
      total_error += GAdb2*t*t;
    } else if ( e <= DELTA )
      credit[k] = 0.0;
    else {
      t = e - DELTA;
      credit[k] = -GI * t;
      total_error += GIdb2*t*t;
    }
    
  }
  
  return( total_error );
}


void user_init()
{
  spiral_set_random_init_range(0.5);

  spiral_set_learning_rate(0.1); 
}
 
