/* 

  ****************   NO WARRANTY  *****************

Since the Aspirin/MIGRAINES system is licensed free of charge,
Russell Leighton and the MITRE Corporation provide absolutley 
no warranty. Should the Aspirin/MIGRAINES system prove defective, 
you must assume the cost of all necessary servicing, repair or correction.
In no way will Russell Leighton or the MITRE Corporation be liable to you for
damages, including any lost profits, lost monies, or other
special, incidental or consequential damages arising out of
the use or inability to use the Aspirin/MIGRAINES system.

  *****************   COPYRIGHT  *******************

This software is the copyright of Russell Leighton and the MITRE Corporation. 
It may be freely used and modified for research and development
purposes. We require a brief acknowledgement in any research
paper or other publication where this software has made a significant
contribution. If you wish to use it for commercial gain you must contact 
The MITRE Corporation for conditions of use. Russell Leighton and 
the MITRE Corporation provide absolutely NO WARRANTY for this software.

   August, 1992 
   Russell Leighton
   The MITRE Corporation
   7525 Colshire Dr.
   McLean, Va. 22102-3481

*/
/**
 ** YACC description of datafile reader
 **/


/*
 * Define parsers stack to be numbers, strings.
 */

%union {
	int i;    /* integer */
	float f;  /* float */
	char *s;  /* string */
}


%{
#include "BpDatafile.h"
#include "globals.h"

 /* initialize files */
extern DFPTR _file_init(LB_PTR (*query_function)(), int noise_type, float mean, float variance);

/* read in a datafile */
extern void  _readfile(DATAFILEPTR datafile, DFPTR df) ;

#define TRUE 1
#define FALSE 0

/* input stream */
static FILE *df_stream;

/* input buffer */
static char yytext[256];
static int yyleng;

DFPTR df = NULL; /* root */

static int error_code = 0;

#define TARGET_LENGTH 50


/* globals */
static int cache_size = 0;      /* number of patterns to cache */
static char *user_label = NULL; /* user supplied label */  
static int target_length = 0;   /* length of user's target vector */
static float target_vector[TARGET_LENGTH];
static int self = FALSE; /* if target is same as data */
static char *target_file = NULL; /* name of target file (type5) */
static int switch_cycle = 1; /* when to switch between files (after how many patterns) */
static float range_threshold, range = -1.0, bias = 0.0, scalar=1.0; /* for rescaling data */
static int normalize = 0; /* for normalizing data */
static int submean = 0; /* for normalizing data */
static float lowclip, highclip; /* for preprocessing data */
static int clear_delays = FALSE; /* TRUE if clear delay buffers before using */
static int random_switch = FALSE;
static int swapbytes= FALSE; /* in case the data is from a machine w/opposite byte order */
static int math=DF_MATH_NONE;       /* for doing math on data points */

/* used for error messages */
int lineno = 1; 
static char *datafile_name;



int yywrap() { return(1); } /* noop */

/*****************************************************/
/* warning                                           */
/*        Print warning message.                     */
/*****************************************************/
static void warning(char *s, char *t)
{
  fprintf(stderr, "\n%s: %s", datafile_name, s);
  if (t != (char *)NULL) fprintf(stderr, "  %s", t);
}

/*****************************************************/
/* yyerror                                           */
/*        Called by yacc for syntax errors.          */
/*****************************************************/
int yyerror (char *s)
{
  char where[256];

  *where = '\0'; /* empty */
  if (yytext != NULL)
    {  char *token = am_alloc_mem(yyleng + 1);

       bcopy(yytext, token, yyleng);
       token[yyleng] = '\0'; /* end of string */
       sprintf(where, "(near \"%s\" somewhere around line %d)\n", yytext, lineno);
       am_free_mem(token);
    }/* end if */

  warning(s, where);
  error_code = 1;

  return(error_code);
}/* end yyerror */  

/*****************************************************/
/*****************************************************/
/*****************************************************/
/*******************    lex    ***********************/



char *newstr (char *name, int length)
{
 char *temp;

 temp = am_alloc_mem(length + 1);
 *(temp + length) = 0;
 bcopy(name, temp, length+1);
 return(temp);
}

static void read_token()
{
  unsigned int c;


  yyleng=0;
  /* skip spaces, tabs , quotes, and newlines  */
  while((c=getc(df_stream)) != ' '
	&& !(iscntrl(c))
	&& c != '"'
	&& c != '{'
	&& c != '}'
	&& c != '('
	&& c != ')'
	&& c != '#'
	&& c != ','
	&& c != EOF) {
    yytext[yyleng++] = c;
  }
  ungetc((unsigned char)c,df_stream);
  yytext[yyleng] = '\0';
}

static void read_quoted_token()
{
  unsigned int c;


  yyleng=0;
  /* end at quote */
  while((c=getc(df_stream)) != '"' && c != EOF) {
    yytext[yyleng++] = c;
  }
  yytext[yyleng] = '\0';
}

static int input_equals(char *str)
{
  register int counter;
  register char *tptr = yytext;

  counter = yyleng;
  if (strlen(str) > counter) counter = strlen(str);
  while(counter--) if (*str++ != *tptr++) return FALSE;;
  return TRUE;

}

#define CONTEXT_STACK_DEPTH 16
#define NO_CONTEXT 0
#define READFILE_CONTEXT 1
#define USEFILE_CONTEXT 2
#define KEYWORD_CONTEXT 3
#define MATH_CONTEXT 4
static int context_stack[CONTEXT_STACK_DEPTH], *cstackptr = context_stack;

#define switch_lex_context(ctx) *cstackptr = ctx;

static void push_lex_context(int new_context)
{
  if (cstackptr == context_stack + CONTEXT_STACK_DEPTH)
    yyerror("\nLex context stack full! (push_lex_context)\n");
  cstackptr++; *cstackptr = new_context;
}

static void pop_lex_context()
{
  if (cstackptr == context_stack)
    yyerror("\nLex context stack would be empty! (pop_lex_context)\n");
  /* pop */
  cstackptr--;
}

static int this_context(int context)
{
  return( *cstackptr == context );
}




/*****************************************************/
/*****************************************************/
/*****************************************************/


/*****************************************************/
/* update_target_vector: Add numbers to target vector. */
/*****************************************************/
static void update_target_vector(float x)
{
  if (target_length == TARGET_LENGTH) yyerror("\nTarget vector too long!");
  target_vector[target_length++] = x;
}/* end update_target_vector */


/*****************************************************/
/* create_datafile: create datafile structure.       */
/*****************************************************/
static void create_datafile(char *filename, int type)
{
  DATAFILEPTR new_datafile;

  /* make new structure */
  new_datafile = (DATAFILEPTR)am_alloc_mem(sizeof(DATAFILE));
  new_datafile->filename = filename;

  new_datafile->type = type;

  new_datafile->usefile = FALSE;

  new_datafile->next = (DATAFILEPTR)NULL;

  new_datafile->size = df->inputs_size;
  new_datafile->xdim = df->inputs_xdim;
  new_datafile->ydim = df->inputs_ydim;

  new_datafile->current_pattern = -1;

  new_datafile->switch_counter = 1;

  new_datafile->switch_cycle = switch_cycle;
  switch_cycle = 1; /* reset */


  new_datafile->range = range;
  range = -1.0; /* reset */
  new_datafile->range_threshold = range_threshold;
  range_threshold = -AM_HUGE_VAL; /* reset */
  new_datafile->bias = bias;
  bias = 0.0; /* reset */
  new_datafile->scalar = scalar;
  scalar = 1.0; /* reset */
  new_datafile->normalize = normalize;
  normalize = 0; /* reset */
  new_datafile->lowclip = lowclip;
  lowclip = -AM_HUGE_VAL; /* reset */
  new_datafile->highclip = highclip;
  highclip = AM_HUGE_VAL; /* reset */
  new_datafile->submean = submean;
  submean = 0; /* reset */
  new_datafile->math= math;
  math = 0; /* reset */
  new_datafile->swapbytes= swapbytes;
  swapbytes = 0; /* reset */

  new_datafile->clear_delays = clear_delays;
  clear_delays = FALSE; /* reset */

  new_datafile->cache_size = cache_size;
  cache_size = 0; /* reset */

  new_datafile->user_label = user_label;
  user_label = (char *)NULL; /* reset */
  
  /* Target-> in .df */
  if (target_length) {

    new_datafile->target.user_target = (float *)am_alloc_mem(target_length * sizeof(float));
    /* copy */
    bcopy((char *)target_vector,
	  (char *)new_datafile->target.user_target,
	  (target_length * sizeof(float)));
    /* reset */
    target_length = 0;

  } else if ( self ) { /* use inputs as targets? */
    
    new_datafile->target.self = 1; self = FALSE;   
    
  } else if (target_file != NULL) {  /* set the target file (type5) */

    new_datafile->target.target_file = target_file; target_file = NULL;

  }/* end else if */

  new_datafile->inputs = (float **)NULL;
  new_datafile->targets = (float **)NULL;

  /* push into list */
  df->n_datafiles += 1;
  if (df->datafiles == (DATAFILEPTR)NULL) {
    df->datafiles = new_datafile;    
    new_datafile->next = (DATAFILEPTR)NULL;
  } else {
    DATAFILEPTR temp;
    temp = df->datafiles;
    df->datafiles = new_datafile;
    new_datafile->next = temp;
  }/* end else */

}/* end create_datafile */

/*****************************************************/
/* find_datafile: Search and return named file     */
/*****************************************************/
DATAFILEPTR find_datafile(char *name)
{
  DATAFILEPTR ptr;

  ptr = df->datafiles;
  while(ptr != (DATAFILEPTR)NULL) {
    if (! ptr->usefile && ! strcmp(ptr->filename, name))
      return ptr;
    else
      ptr = ptr->next;
  }

  yyerror("Undefined file name for UseFile!");

  return ( NULL );

}/* end find_datafile */

/*****************************************************/
/* use_datafile:   Just reference a file             */
/*****************************************************/
static void use_datafile(char *name)
{
  DATAFILEPTR referenced_datafile, new_datafile;

  /* find name in list */
  referenced_datafile = find_datafile(name);

  /* make new structure */
  new_datafile = (DATAFILEPTR)am_alloc_mem(sizeof(DATAFILE));

  new_datafile->filename = name;

  /* mark so it won't be read */
  new_datafile->usefile = TRUE;

  /* push into list */
  df->n_datafiles += 1;
  {
    DATAFILEPTR temp;
    temp = df->datafiles;
    df->datafiles = new_datafile;
    new_datafile->next = temp;
  }/* end block */

}/* end use_datafile */

/*****************************************************/
/* next_datafile: Goto next datafile. */
/*****************************************************/
static void next_datafile()
{
    df->current_datafile = df->current_datafile->next;
}/* end next_datafile */

/*****************************************************/
/* random_datafile: random next datafile. */
/*****************************************************/
static void random_datafile()
{
    df->current_datafile = *(df->datafile_table +
			       ( (int) (AM_RANDOM() * df->n_datafiles) ));
}/* end next_datafile */


%}

/*
 * Enumerate all tokens from yylex.
 */

/* This is used to reset the line counter */
%token NEWLINE

/* The following are the keyword tokens */
%token <f> NUMBER                 /* real */
%token <s> NAME                   /* string */

%token LCP                        /* left curly paren */
%token RCP                        /* right curly paren */
%token RP                         /* right paren */
%token LP                         /* left paren */

%token READFILE                   /* read file */
%token USEFILE                    /* reference a file */
%token RANDOM_SWITCH              /* randomly switch files */

%token ASCII                      /* file type */
%token MATLAB                     /* file type */
%token TYPE1                      /* file type */
%token TYPE2                      /* file type */
%token TYPE3                      /* file type */
%token TYPE4                      /* file type */
%token TYPE5                      /* file type */

%token CACHE                      /* key word */
%token LABEL                      /* key word */
%token TARGET                     /* key word */
%token SELF                       /* argument of TARGET */
%token SWITCH                     /* key word */

%token RANGE                      /* key word */
%token NORMALIZE                  /* key word */
%token LOWCLIP                    /* key word */
%token HIGHCLIP                   /* key word */
%token SUBMEAN                    /* key word */
%token SCALE                      /* key word */
%token BIAS                       /* key word */
%token CLEAR_DELAYS               /* key word */
%token SWAPBYTES                  /* key word */
%token MATH                       /* key word */
%token <i> MATH_FUNCTION          /* key word set */

/*
 * Sub-expression (non-terminal) tokens.
 */

%type <i> type
%type <f> real
%type <i> integer
%type <i> optional_integer
%%


/*****************************************************/
/* Root production
/*****************************************************/
datafile: datafile_exp
         | datafile datafile_exp
         ;

/* these are the currently supported datafile expressions */
datafile_exp: files    /* new files */
              | newlines   /* cpp statements */
              ;

/* read a list of commands */
files: READFILE LP NAME type keys RP { create_datafile($3, $4); }
           | USEFILE LP NAME RP { use_datafile($3); }
           | RANDOM_SWITCH { random_switch = TRUE; }
           ;


/* what type? */
type : ASCII { $$ = ASCII_FILE; }
       | MATLAB { $$ = MATLAB_FILE; }
       | TYPE1 { $$ = TYPE1_FILE; }
       | TYPE2 { $$ = TYPE2_FILE; }
       | TYPE3 { $$ = TYPE3_FILE; }
       | TYPE4 { $$ = TYPE4_FILE; }
       | TYPE5 { $$ = TYPE5_FILE; }
       ;

/* list of key words */
keys : key
       | keys key

/* key words */
key: LP LABEL NAME RP { user_label = $3; }
      | LP TARGET vector RP
      | LP TARGET SELF RP { self = TRUE ; }
      | LP TARGET NAME RP { target_file = $3; }
      | LP CACHE integer RP { if ($3 < 1) yyerror("Cache size must be postive!"); cache_size = $3; }
      | LP SWITCH optional_integer RP { switch_cycle = $3; }
      | LP RANGE real real RP { if ($3 <= 0.0) yyerror("Range must be postive!");
                                range = $3; range_threshold = $4; }
      | LP RANGE real RP { if ($3 <= 0.0) yyerror("Range must be postive!"); range = $3; }
      | LP BIAS real RP { bias = $3; }
      | LP SCALE real RP { scalar = $3; }
      | LP NORMALIZE RP { normalize = 1; }
      | LP LOWCLIP real RP { lowclip = $3; }
      | LP HIGHCLIP real RP { highclip = $3; }
      | LP SWAPBYTES RP { swapbytes = 1; }
      | LP MATH MATH_FUNCTION RP { math = $3; }
      | LP SUBMEAN RP { submean = 1; }
      | LP CLEAR_DELAYS RP {clear_delays = TRUE; }
      |
      ;

/* list of numbers */
 vector: real { update_target_vector($1); }
         | vector real { update_target_vector($2); }
         ;

/*****************************************************/
/* Update the line counter and current file          */
/* (used to parse cpp output)                        */
/*****************************************************/
newlines: newline
          | newlines newline
          ;

newline: NEWLINE integer NAME { lineno = $2;
 				datafile_name = $3;
			      } 
         | NEWLINE integer  NAME integer { 
	                                   lineno = $2;
 				           datafile_name = $3;
			                 }
         | NEWLINE NAME integer NAME { lineno = $3;
                                       datafile_name = $4;
                                     }
         | NEWLINE NAME NAME {}

         | NEWLINE integer { lineno = $2; }

         ;

/* numbers */
real: NUMBER { $$ = $1; }
     ;

optional_integer: integer { $$ = $1; }
                |  { $$ = 1; /* default */}
                ;

integer: NUMBER {$$ = (int)$1; }
         ; 
%%
  

int yylex()
{
  unsigned int c;

  more :


  /* skip spaces and tabs */
  while((c=getc(df_stream)) == ' ' || c == '\t') ;

  if (c == '\n') {  ++lineno; goto more ; }

  if (iscntrl(c)) goto more ;
  
  if (c == EOF) return 0;

  if (c == '#') return(NEWLINE);

  if (c == '"') { /* quoted string */
    read_quoted_token();
    yylval.s = newstr(yytext, yyleng);
    return(NAME);
  }

  if (c == '(') {
    if (this_context(KEYWORD_CONTEXT))
      push_lex_context(KEYWORD_CONTEXT);
    return(LP);
  }/* end if */

  if (c == ')') {
    pop_lex_context();
    return(RP);
  }/* end if */


  /* ok, done with the single character stuff */
  ungetc((unsigned char)c, df_stream);
  read_token();

  if (isdigit((unsigned int)yytext[0]) ||
      (yyleng > 1 &&
       (yytext[0] == '-' || yytext[0] == '+' || yytext[0] == '.') &&
       isdigit((unsigned int)yytext[1])) ||
      (yyleng > 2 &&
       (yytext[0] == '-' || yytext[0] == '+') &&
       yytext[1] == '.' &&
       isdigit((unsigned int)yytext[2]))) { /* number */
    sscanf((char *)yytext, "%f", &(yylval.f)); 
    return(NUMBER);
  }/* end if number */ 

  if (this_context(NO_CONTEXT)) { /* global context */
    if (input_equals("RandomSwitch")) return(RANDOM_SWITCH);
  }

  if (this_context(NO_CONTEXT)) { /* global context */
    if (input_equals("UseFile")) {
      push_lex_context(USEFILE_CONTEXT);
      return(USEFILE);
    }
  }

  if (this_context(NO_CONTEXT)) { /* global context */
    if (input_equals("ReadFile")) {
      push_lex_context(READFILE_CONTEXT);
      return(READFILE);
    }
  }

  if (this_context(READFILE_CONTEXT)) {

    /* types */
    
    if (input_equals("Ascii")) {
      switch_lex_context(KEYWORD_CONTEXT);
      return(ASCII);
    }
    
    if (input_equals("Matlab")) {
      switch_lex_context(KEYWORD_CONTEXT);
      return(MATLAB);
    }
    
    if (input_equals("Type1")) {
      switch_lex_context(KEYWORD_CONTEXT);
      return(TYPE1);
    }

    if (input_equals("Type2")) {
      switch_lex_context(KEYWORD_CONTEXT);
      return(TYPE2);
    }

    if (input_equals("Type3")) {
      switch_lex_context(KEYWORD_CONTEXT);
      return(TYPE3);
    }

    if (input_equals("Type4")) {
      switch_lex_context(KEYWORD_CONTEXT);
      return(TYPE4);
    }

    if (input_equals("Type5")) {
      switch_lex_context(KEYWORD_CONTEXT);
      return(TYPE5);
    }
    
  }

  if (this_context(KEYWORD_CONTEXT)) {

    if (input_equals("Math->")) {
      push_lex_context(MATH_CONTEXT);
      return(MATH);
    }

    if (input_equals("ClearDelays->")) return(CLEAR_DELAYS);

    if (input_equals("SubMean->")) return(SUBMEAN);

    if (input_equals("Normalize->")) return(NORMALIZE);

    if (input_equals("LowClip->")) return(LOWCLIP);

    if (input_equals("HighClip->")) return(HIGHCLIP);

    if (input_equals("Scale->")) return(SCALE);

    if (input_equals("Bias->")) return(BIAS);

    if (input_equals("Range->")) return(RANGE);

    if (input_equals("SwapBytes->")) return(SWAPBYTES);

    if (input_equals("Switch->")) return(SWITCH);

    if (input_equals("Target->")) return(TARGET);

    if (input_equals("Label->")) return(LABEL);

    if (input_equals("Cache->")) return(CACHE);

    if (input_equals("All")) { yylval.f = -1.0 ; return(NUMBER); }

    if (input_equals("Self")) return(SELF);

  }

  if (this_context(MATH_CONTEXT)) {

    if (input_equals("abs[x]"))  yylval.i = DF_MATH_FABS;
    
    else if (input_equals("square[x]"))  yylval.i = DF_MATH_SQUARE;

    else if (input_equals("log10[x]"))  yylval.i = DF_MATH_LOG10;

    else if (input_equals("log10[x+1]"))  yylval.i = DF_MATH_PSEUDO_LOG10;

    else fprintf(stderr, "\nMath->: Unknown function!\n");

    pop_lex_context();

    return(MATH_FUNCTION);
  }


  /* names */
  {
    yylval.s = newstr(yytext, yyleng);
    return(NAME);
  }  

}/* end yylex */


int _read_df (char *file)
{
  char *cpp_file;

  /******************************************/
  /* some basic init */
  lowclip = -AM_HUGE_VAL;
  highclip = AM_HUGE_VAL;
  range_threshold = -AM_HUGE_VAL;
  datafile_name = file;
  
  /******************************************/
  /* run cpp on datafile */
  cpp_file = am_cpp(file,0);

  /******************************************/
  /* open  input from cpp */
  df_stream = am_fopen(cpp_file, "r");
  /* return on error */
  if (df_stream == NULL)  {
    fprintf(stderr, "\nUnable to open %s", cpp_file);
    am_perror("Reading .df file");
    return(1);
  }

  /******************************************/
  
  /******************************************/
  /* parse storing info in global variables */
  yyparse();
  if (error_code) return(error_code);
  /******************************************/

  /******************************************/
  /* close the stream */
  am_fclose(df_stream);
  /******************************************/

  /******************************************/
  /* read in the data                       */
  /******************************************/
  {
    DATAFILEPTR files = df->datafiles;

    if (files == (DATAFILEPTR)NULL) {
      fprintf(stderr, "\nNo data files!\n");
      return(1);
    }
    
    /* loop through each structure...read */
    while(files != (DATAFILEPTR)NULL) {
      if (!files->usefile) _readfile(files, df);
      files = files->next;
    }/* end while */

    
    /* set the first one */
    df->current_datafile = df->datafiles;

  }/* end block */
  
  /******************************************/
  /* Reference files w/UseFile              */
  /******************************************/
  {
    DATAFILEPTR files = df->datafiles;

    /* loop through each structure...reference (from UseFile) */
    while(files != (DATAFILEPTR)NULL) {

      if (files->usefile) { /* just a reference, don't read */
	DATAFILEPTR tmp_next = files->next; /* hold on */

	/* copy contents */
	*files = *(find_datafile(files->filename));

	/* fix list */
	files->next = tmp_next;
	files->usefile = TRUE;

      }/* end if */

      files = files->next;
    }/* end while */

  }/* end block */

  /******************************************/  
  /* preprocess                             */
  /******************************************/  
  {
    extern void _preprocess_df(DATAFILEPTR datafile); /* from preprocess.c */
    DATAFILEPTR files = df->datafiles;
    /* loop through each structure... */
    while(files != (DATAFILEPTR)NULL) {

      _preprocess_df(files);
      files = files->next;
    }/* end while */
  }/* end block */
  /******************************************/  

  /******************************************/
  /* setup for the next_file function       */
  /******************************************/
  {
    /* set the next file function */
    if (random_switch) { /* random */
      DATAFILEPTR files = df->datafiles, *file_ptr;
      
      df->datafile_table = (DATAFILEPTR *)am_alloc_mem(df->n_datafiles * sizeof(DATAFILEPTR));
      file_ptr = df->datafile_table;
      
      /* loop through each structure...set into table */
      while(files != (DATAFILEPTR)NULL) {
	*file_ptr++ = files;
	files = files->next;
      }/* end while */
      
      df->next_datafile = random_datafile;
      
    } else { /* cyclic */
      
      /* set function */
      df->next_datafile = next_datafile;
      
    }/* end else */
    
  }/* end block */



  /******************************************/
  /* make circular list                     */
  /******************************************/
  {
    DATAFILEPTR files = df->datafiles, last_file=df->datafiles;

    /* loop through each structure...find the end */
    while(files != (DATAFILEPTR)NULL) {
      last_file = files;
      files = files->next;
    }/* end while */
    
    /* make circular list */
    if (df->datafiles != (DATAFILEPTR)NULL) 
      last_file->next = df->datafiles;
    
  }/* end block */
  
  /******************************************/
  /* remove cpp file */
  am_remove_cpp_file(cpp_file);
  /******************************************/

  return(error_code);
}




/*****************************************************/
/* loaddata:       file = datafile specification from
                   which datafiles are read.
		   query_function = function
		   to interogate the network.
		   noise_type = type of noise to
		   corrupt input (if any)
		   mean = mean of noise distribution.
		   variance = variance of noise distribution.
 */
/*****************************************************/
int loaddata(char *file, LB_PTR (*query_function)(), int noise_type, float mean, float variance)
{
  
  /******************************************/
  /* make the root */
  df = _file_init(query_function, noise_type, mean, variance);

  /******************************************/
  /* read the .df */
  return( _read_df(file) );
  /******************************************/
}
